#!/usr/bin/env python

# Copyright (c) 2020, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/spherical_functions/blob/master/LICENSE>
"""Script and/or library to calculate optimal translation and boost from Horizons.h5

See docstrings below for more detail.

To use this function from another script, simply do something like the following:

    from scri.SpEC.com_motion import com_motion, estimate_avg_com_motion
    x_0, v_0, t_0 = estimate_avg_com_motion('path/to/Horizons.h5')

Other options may be passed to the `estimate_avg_com_motion` function; see its docstring for details.

"""


def com_motion(path_to_horizons_h5, path_to_matter_h5=None, m_A=None, m_B=None):
    """Calculate the center-of-mass motion from Horizons.h5 and/or Matter.h5

    Note that no input checking is done here.  If the data sets are missing, incomplete, or corrupted, you should
    expect to get exceptions or silently incorrect data.

    Parameters
    ----------
    path_to_horizons_h5 : string, optional [default is 'Horizons.h5']
        Absolute or relative path to 'Horizons.h5'
    path_to_matter_h5: None or string [default is None]
        Absolute or relative path to 'Matter.h5' for neutron star trajectories.
        If None, we assume that only black holes are involved, so that both A
        and B trajectories will be in Horizons.h5.
    m_A: None or float [default is None]
        If None, the mass will be read from Horizons.h5 (for BHs) or Matter.h5
        (for NSs).
    m_B: None or float [default is None]
        If None, the mass will be read from Horizons.h5 (for BHs) or Matter.h5
        (for NSs).

    Returns
    -------
    t : float array
        These are simply the coordinate times
    com : 3xN float array
        Position of CoM at each instant of time

    """
    import numpy as np
    import h5py

    if path_to_matter_h5 is None:
        with h5py.File(path_to_horizons_h5, "r") as horizons:
            t = horizons["AhA.dir/ChristodoulouMass.dat"][:, 0]
            m_A = horizons["AhA.dir/ChristodoulouMass.dat"][:, 1]
            x_A = horizons["AhA.dir/CoordCenterInertial.dat"][:, 1:]
            m_B = horizons["AhB.dir/ChristodoulouMass.dat"][:, 1]
            x_B = horizons["AhB.dir/CoordCenterInertial.dat"][:, 1:]
            m = m_A + m_B
            CoM = ((m_A[:, np.newaxis] * x_A) + (m_B[:, np.newaxis] * x_B)) / m[:, np.newaxis]
    else:
        # NOTE: Currently only supports BHNS, not NSNS
        with h5py.File(path_to_horizons_h5, "r") as horizons:
            t = horizons["AhA.dir/ChristodoulouMass.dat"][:, 0]
            if m_A is None:
                m_A = horizons["AhA.dir/ChristodoulouMass.dat"][:, 1]
            else:
                m_A = np.atleast_1d(m_A)
            x_A = horizons["AhA.dir/CoordCenterInertial.dat"][:, 1:]
        with h5py.File(path_to_matter_h5, "r") as matter:
            if m_B is None:
                m_B = matter["RestMass.dat"][:, 1]
            else:
                m_B = np.atleast_1d(m_B)
            x_B = matter["InertialCenterOfMass.dat"][:, 1:]
        m = m_A + m_B
        CoM = ((m_A[:, np.newaxis] * x_A) + (m_B[:, np.newaxis] * x_B)) / m[:, np.newaxis]

        # The waveform will be in units of the total mass, so we need to convert for compatibility
        t /= m
        CoM /= m[:, np.newaxis]

    return t, CoM


def estimate_avg_com_motion(
    path_to_horizons_h5="Horizons.h5",
    skip_beginning_fraction=0.01,
    skip_ending_fraction=0.10,
    plot=False,
    fit_acceleration=False,
    path_to_matter_h5=None,
    m_A=None,
    m_B=None,
):
    """Calculate optimal translation and boost from Horizons.h5

    This returns the optimal initial position and velocity such that the CoM is
    best approximated as having these initial values.  If the coordinate system
    is transformed by these values, the new CoM motion will be as close to the
    origin as possible (in the sense of squared distance from the origin
    integrated over time).

    The translation to be applied to the data should be calculated given the
    values returned by this function as

        np.array([x_i+v_i*t_j+0.5*a_i*t_j**2 for t_j in t])


    Parameters
    ----------
    path_to_horizons_h5: string, optional [default is 'Horizons.h5']
        Absolute or relative path to 'Horizons.h5'
    skip_beginning_fraction : float, optional
        Exclude this portion from the beginning of the data.  Note that this is
        a fraction, rather than a percentage.  The default value is 0.01,
        meaning the first 1% of the data will be ignored.
    skip_ending_fraction : float, optional
        Exclude this portion from the end of the data.  Note that this is a
        fraction, rather than a percentage.  The default value is 0.10, meaning
        the last 10% of the data will be ignored.
    plot : bool, optional
        If True, save plot showing CoM tracks before and after offset and
        translation, to file `CoM_before_and_after_translation.pdf` in the same
        directory as Horizons.h5.  Default: False.
    fit_acceleration: bool, optional
        If True, allow for an acceleration in the fit, and return as third
        parameter.  Default: False.
    path_to_matter_h5: None or string [default is None]
        Absolute or relative path to 'Matter.h5' for neutron star trajectories.
        If None, we assume that only black holes are involved, so that both A
        and B trajectories will be in Horizons.h5.
    m_A: None or float [default is None]
        If None, the mass will be read from Horizons.h5 (for BHs) or Matter.h5
        (for NSs).
    m_B: None or float [default is None]
        If None, the mass will be read from Horizons.h5 (for BHs) or Matter.h5
        (for NSs).

    Returns
    -------
    x_i : length-3 array of floats
        Best-fit initial position of the center of mass
    v_i : length-3 array of floats
        Best-fit initial velocity of the center of mass
    a_i : length-3 array of floats
        Best-fit initial acceleration of the center of mass [only if
        `fit_acceleration=True` is in input arguments]
    t_i : float
        Initial time used.  This is determined by the `skip_beginning_fraction`
        input parameter.
    t_f : float
        Final time used.  This is determined by the `skip_ending_fraction`
        input parameter.

    """
    import os.path
    import numpy as np
    from scipy.integrate import simps

    t, com = com_motion(path_to_horizons_h5, path_to_matter_h5, m_A, m_B)

    # We will be skipping the beginning and end of the data;
    # this gives us the initial and final indices
    t_i, t_f = t[0] + (t[-1] - t[0]) * skip_beginning_fraction, t[-1] - (t[-1] - t[0]) * skip_ending_fraction
    i_i, i_f = np.argmin(np.abs(t - t_i)), np.argmin(np.abs(t - t_f))

    # Find the optimum analytically
    CoM_0 = simps(com[i_i : i_f + 1], t[i_i : i_f + 1], axis=0)
    CoM_1 = simps((t[:, np.newaxis] * com)[i_i : i_f + 1], t[i_i : i_f + 1], axis=0)
    if fit_acceleration:
        CoM_2 = simps((t[:, np.newaxis] ** 2 * com)[i_i : i_f + 1], t[i_i : i_f + 1], axis=0)
        x_i = (
            3
            * (
                CoM_0
                * (3 * t_f ** 4 + 12 * t_f ** 3 * t_i + 30 * t_f ** 2 * t_i ** 2 + 12 * t_f * t_i ** 3 + 3 * t_i ** 4)
                - 12 * CoM_1 * (t_f + t_i) * (t_f ** 2 + 3 * t_f * t_i + t_i ** 2)
                + CoM_2 * (10 * t_f ** 2 + 40 * t_f * t_i + 10 * t_i ** 2)
            )
            / (t_f - t_i) ** 5
        )
        v_i = (
            12
            * (
                -3 * CoM_0 * (t_f + t_i) * (t_f ** 2 + 3 * t_f * t_i + t_i ** 2)
                + CoM_1 * (16 * t_f ** 2 + 28 * t_f * t_i + 16 * t_i ** 2)
                + CoM_2 * (-15 * t_f - 15 * t_i)
            )
            / (t_f - t_i) ** 5
        )
        a_i = (
            60
            * (CoM_0 * (t_f ** 2 + 4 * t_f * t_i + t_i ** 2) + CoM_1 * (-6 * t_f - 6 * t_i) + 6 * CoM_2)
            / (t_f - t_i) ** 5
        )
    else:
        x_i = 2 * (CoM_0 * (2 * t_f ** 3 - 2 * t_i ** 3) + CoM_1 * (-3 * t_f ** 2 + 3 * t_i ** 2)) / (t_f - t_i) ** 4
        v_i = 6 * (CoM_0 * (-t_f - t_i) + 2 * CoM_1) / (t_f - t_i) ** 3
        a_i = 0.0

    # If desired, save the plots
    if plot:
        import matplotlib as mpl

        mpl.use("Agg")  # Must come after importing mpl, but before importing plt
        import matplotlib.pyplot as plt

        directory = os.path.dirname(os.path.abspath(path_to_horizons_h5))
        try:
            from subprocess import check_output

            SXS_BBH = check_output(
                "awk '/alternative-names/{{print $3}}' " + os.path.join(directory, "metadata.txt"), shell=True
            )
            if SXS_BBH:
                SXS_BBH = "\n" + SXS_BBH.strip()
        except:
            SXS_BBH = ""
        delta_x = np.array([x_i + v_i * t_j + 0.5 * a_i * t_j ** 2 for t_j in t])
        comprm = com - delta_x
        max_displacement = np.linalg.norm(delta_x, axis=1).max()
        max_d_color = min(1.0, 10 * max_displacement)
        fig = plt.figure(figsize=(10, 7))
        plt.plot(t, com, alpha=0.25, lw=1.5)
        plt.gca().set_prop_cycle(None)
        lineObjects = plt.plot(t, comprm, lw=2)
        plt.xlabel(r"Coordinate time")
        plt.ylabel(r"CoM coordinate values")
        plt.legend(iter(lineObjects), ("x", "y", "z"), loc="upper left")
        fig.text(
            0.5,
            0.91,
            directory
            + SXS_BBH
            + "\n$x_i$ = [{}]".format(", ".join([str(tmp) for tmp in x_i]))
            + "\n$v_i$ = [{}]".format(", ".join([str(tmp) for tmp in v_i])),
            fontsize=8,
            ha="center",
        )
        fig.text(
            0.004,
            0.004,
            str(max_displacement),
            fontsize=24,
            ha="left",
            va="bottom",
            bbox=dict(facecolor=mpl.cm.jet(max_d_color), alpha=max_d_color, boxstyle="square,pad=0.2"),
        )
        fig.savefig(os.path.join(os.path.dirname(path_to_horizons_h5), "CoM_before_and_after_transformation.pdf"))
        plt.close()

    print("Optimal x_i: [{}, {}, {}]".format(*x_i))
    print("Optimal v_i: [{}, {}, {}]".format(*v_i))
    if fit_acceleration:
        print("Optimal a_i: [{}, {}, {}]".format(*a_i))
    print(f"t_i, t_f: {t_i}, {t_f}")

    if fit_acceleration:
        return x_i, v_i, a_i, t_i, t_f
    else:
        return x_i, v_i, t_i, t_f


def remove_avg_com_motion(
    path_to_waveform_h5="rhOverM_Asymptotic_GeometricUnits.h5/Extrapolated_N2.dir",
    path_to_horizons_h5=None,
    skip_beginning_fraction=0.01,
    skip_ending_fraction=0.10,
    plot=False,
    file_write_mode="w",
    path_to_matter_h5=None,
    m_A=None,
    m_B=None,
    file_format="NRAR",
):
    """Rewrite waveform data in center-of-mass frame

    This simply uses `estimate_avg_com_motion`, and then transforms to that
    frame as appropriate.  Most of the options are simply passed to that
    function.  Note, however, that the path to the Horizons.h5 file defaults to
    the directory of the waveform H5 file.

    Additional parameters
    ---------------------
    path_to_waveform_h5 : str, optional
        Absolute or relative path to SpEC waveform file, including the
        directory within the H5 file, if appropriate.  Default value is
        'rhOverM_Asymptotic_GeometricUnits.h5/Extrapolated_N2.dir'.
    path_to_horizons_h5: None or string [default is None]
        Absolute or relative path to 'Horizons.h5'.  If None, this will be
        inferred from the waveform path.
    path_to_matter_h5: None or string [default is None]
        Absolute or relative path to 'Matter.h5' for neutron star trajectories.
        If None, this will be inferred from the waveformpath.  If the file does
        not exist, we assume that only black holes are involved, so that both A
        and B trajectories will be in Horizons.h5.
    m_A: None or float [default is None]
        If None and there is no Matter.h5 file, the mass will be read from
        Horizons.h5; otherwise, the mass will be read from the metadata
        `reference_mass1`.
    m_B: None or float [default is None]
        If None and there is no Matter.h5 file, the mass will be read from
        Horizons.h5; otherwise, the mass will be read from the metadata
        `reference_mass2`.
    file_format: 'NRAR' or 'RPXMB' [default is 'NRAR']
        The file format of the waveform data H5 file. The 'NRAR' format is the
        default file format found in the SXS Catalog. The 'RPXMB' format is data
        compressed with the rotating_paired_xor_multishuffle_bzip2 scheme.

    Returns
    -------
    w_m : WaveformModes object
        This is the transformed object in the new frame

    """
    import os.path
    import pathlib
    import re
    import numpy as np
    from .file_io import read_from_h5, write_to_h5, rotating_paired_xor_multishuffle_bzip2
    from .. import h, hdot, psi4, psi3, psi2, psi1, psi0

    directory = os.path.dirname(os.path.abspath(path_to_waveform_h5.split(".h5", 1)[0] + ".h5"))
    subdir = os.path.basename(path_to_waveform_h5.split(".h5", 1)[1])

    # Read the waveform data in
    if file_format.lower() == "nrar":
        w_m = read_from_h5(path_to_waveform_h5)
    elif file_format.lower() == "rpxmb" or file_format.lower() == "rpxm":
        w_m = rotating_paired_xor_multishuffle_bzip2.load(path_to_waveform_h5)[0].to_inertial_frame()
    else:
        raise ValueError(f"File format {file_format} not recognized. Must be either 'NRAR' or 'RPXMB'.")

    if path_to_horizons_h5 is None:
        path_to_horizons_h5 = os.path.join(directory, "Horizons.h5")

    if path_to_matter_h5 is None:
        tmp = pathlib.Path(directory) / "Matter.h5"
        if tmp.exists():
            path_to_matter_h5 = str(tmp)
            if m_A is None:
                m_A = w_m.metadata.reference_mass1
            if m_B is None:
                m_B = w_m.metadata.reference_mass2

    # Get the CoM motion from Horizons.h5
    x_0, v_0, t_0, t_f = estimate_avg_com_motion(
        path_to_horizons_h5=path_to_horizons_h5,
        skip_beginning_fraction=skip_beginning_fraction,
        skip_ending_fraction=skip_ending_fraction,
        plot=plot,
        path_to_matter_h5=path_to_matter_h5,
        m_A=m_A,
        m_B=m_B,
    )

    # Set up the plot and plot the original data
    if plot:
        import matplotlib as mpl

        mpl.use("Agg")  # Must come after importing mpl, but before importing plt
        import matplotlib.pyplot as plt
        import scri.plotting

        try:
            if isinstance(w_m.metadata.alternative_names, list):
                SXS_BBH = "\n" + ", ".join(w_m.metadata.alternative_names)
            else:
                SXS_BBH = "\n" + w_m.metadata.alternative_names
        except:
            SXS_BBH = ""
        t_merger = w_m.max_norm_time() - 300.0
        t_ringdown = w_m.max_norm_time() + 100.0
        t_final = w_m.t[-1]
        delta_x = np.array([x_0 + v_0 * t_j for t_j in w_m.t])
        max_displacement = np.linalg.norm(delta_x, axis=1).max()
        max_d_color = min(1.0, 9 * max_displacement)
        LM_indices1 = [[2, 2], [2, 1], [3, 3], [3, 1], [4, 3]]
        LM_indices1 = [[ell, m] for ell, m in LM_indices1 if [ell, m] in w_m.LM.tolist()]
        indices1 = [(ell * (ell + 1) - 2 ** 2 + m) for ell, m in LM_indices1]
        LM_indices2 = [[2, 2], [3, 2], [4, 4], [4, 2]]
        LM_indices2 = [[ell, m] for ell, m in LM_indices2 if [ell, m] in w_m.LM.tolist()]
        indices2 = [(ell * (ell + 1) - 2 ** 2 + m) for ell, m in LM_indices2]
        fig1 = plt.figure(1, figsize=(10, 7))
        plt.gca().set_xscale("merger_zoom", t_merger=t_merger, t_ringdown=t_ringdown, t_final=t_final)
        lines1 = plt.semilogy(w_m.t, abs(w_m.data[:, indices1]), alpha=0.35, lw=1.5)
        fig2 = plt.figure(2, figsize=(10, 7))
        plt.gca().set_xscale("merger_zoom", t_merger=t_merger, t_ringdown=t_ringdown, t_final=t_final)
        lines2 = plt.semilogy(w_m.t, abs(w_m.data[:, indices2]), alpha=0.35, lw=1.5)

    # Import auxilliary waveforms if we are transforming Psi3, Psi2, Psi1, or Psi0. To apply a CoM
    # correction to these data types, information is required from all higher ranked Weyl scalars,
    # e.g. Psi2 requires information from Psi3 and Psi4.
    aux_waveforms = {}
    if w_m.dataType not in [h, hdot, psi4]:
        file_prefix = ["rMPsi4", "r2Psi3", "r3Psi2OverM", "r4Psi1OverM2"]
        regex_template = r"""(r2News|rhOverM|rMPsi4|r2Psi3|r3Psi2OverM|r4Psi1OverM2|r5Psi0OverM3)"""
        for i in range(5 - w_m.dataType):
            match = re.search(regex_template, path_to_waveform_h5)
            # Assume the auxiliary waveforms are in the same directory as the waveform to be transformed
            aux_path = path_to_waveform_h5.replace(match.group(1), file_prefix[i])
            try:
                if file_format.lower() == "nrar":
                    aux_waveforms[f"psi{4-i}_modes"] = read_from_h5(aux_path)
                elif file_format.lower() == "rpxmb" or file_format.lower() == "rpxm":
                    aux_waveforms[f"psi{4-i}_modes"] = rotating_paired_xor_multishuffle_bzip2.load(aux_path)[0]
                    aux_waveforms[f"psi{4-i}_modes"].to_inertial_frame()
            except (FileNotFoundError, OSError):
                raise FileNotFoundError(
                    f"A transformation of {w_m.data_type_string} requires waveform data from Psi{4-i}, "
                    "but the file {aux_path} could not be found."
                )

    # Transform the mode data
    w_m = w_m.transform(space_translation=x_0, boost_velocity=v_0, **aux_waveforms)

    # Write the data to the new file
    if file_format.lower() == "nrar":
        path_to_new_waveform_h5 = re.sub(
            w_m.descriptor_string + "_",
            "",  # Remove 'rhOverM_', 'rMPsi4_', or whatever
            path_to_waveform_h5.replace(".h5", "_CoM.h5", 1),  # Add '_CoM' once
            flags=re.I,
        )  # Ignore case of 'psi4'/'Psi4', etc.
        write_to_h5(
            w_m,
            path_to_new_waveform_h5,
            file_write_mode=file_write_mode,
            attributes={"space_translation": x_0, "boost_velocity": v_0},
        )
    w_m.boost_velocity = v_0
    w_m.space_translation = x_0
    if file_format.lower() == "rpxmb" or file_format.lower() == "rpxm":
        path_to_new_waveform_h5 = path_to_waveform_h5.replace(".h5", "_CoM.h5", 1)
        rotating_paired_xor_multishuffle_bzip2.save(w_m, path_to_new_waveform_h5)

    # Finish by plotting the new data and save to PDF
    if plot:
        plt.figure(1)
        for line, index, (ell, m) in zip(lines1, indices1, LM_indices1):
            plt.semilogy(w_m.t, abs(w_m.data[:, index]), color=plt.getp(line, "color"), lw=1.5, label=f"({ell}, {m})")
        plt.figure(2)
        for line, index, (ell, m) in zip(lines2, indices2, LM_indices2):
            plt.semilogy(w_m.t, abs(w_m.data[:, index]), color=plt.getp(line, "color"), lw=1.5, label=f"({ell}, {m})")
        for fig, num in [(fig1, 1), (fig2, 2)]:
            plt.figure(num)
            plt.axvline(t_merger, color="black", lw=2, alpha=0.125)
            plt.axvline(t_ringdown, color="black", lw=2, alpha=0.125)
            plt.xlabel(r"Coordinate time")
            plt.ylabel(r"Mode amplitudes")
            plt.xlim(0.0, w_m.t[-1])
            plt.ylim(1e-6, 0.6)
            plt.grid(axis="y")
            fig.text(
                0.5,
                0.91,
                os.path.abspath(path_to_waveform_h5)
                + SXS_BBH
                + "\n$x_0$ = [{}]".format(", ".join([str(tmp) for tmp in x_0]))
                + "\n$v_0$ = [{}]".format(", ".join([str(tmp) for tmp in v_0])),
                fontsize=8,
                ha="center",
            )
            fig.text(
                0.004,
                0.004,
                str(max_displacement),
                fontsize=24,
                ha="left",
                va="bottom",
                bbox=dict(facecolor=mpl.cm.jet(max_d_color), alpha=max_d_color, boxstyle="square,pad=0.2"),
            )
            plt.legend(loc="upper left", framealpha=0.9)
            fig.savefig(os.path.join(directory, "Modes_{}_{}_{}.pdf".format(w_m.descriptor_string, subdir[:-4], num)))
            plt.close()

    return w_m


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="Calculate optimal translation and boost from Horizons.h5",
        epilog=(
            "This simply attempts to minimize the squared distance of the CoM from the origin by applying some "
            "constant offset and linear-in-time translation.  Because it uses optimization, it may be sensitive "
            "to the initial guess and may not converge."
        ),
    )
    parser.add_argument(
        "path_to_horizons_h5", default="Horizons.h5", nargs="?", action="store", help="path to the Horizons.h5 file"
    )
    parser.add_argument(
        "--skip_beginning_fraction",
        default=0.01,
        nargs=1,
        type=float,
        action="store",
        help="optional initial portion of the data to ignore",
    )
    parser.add_argument(
        "--skip_ending_fraction",
        default=0.10,
        nargs=1,
        type=float,
        action="store",
        help="optional final portion of the data to ignore",
    )
    parser.add_argument(
        "--plot",
        action="store_true",
        help="make plots as `CoM_before_and_after_transformation.pdf` in same directory as Horizons.h5",
    )
    args = parser.parse_args()

    estimate_avg_com_motion(
        path_to_horizons_h5=args.path_to_horizons_h5,
        skip_beginning_fraction=args.skip_beginning_fraction,
        skip_ending_fraction=args.skip_ending_fraction,
        plot=args.plot,
    )



import copy
import numpy as np
from scipy.interpolate import CubicSpline
import spherical_functions


class ModesTimeSeries(spherical_functions.Modes):
    """Object to store SWSH modes as functions of time

    This class subclasses the spinsfast.Modes class, but also tracks corresponding time values,
    allowing this class to have extra methods for interpolation, as well as differentiation and
    integration in time.

    NOTE: The time array is not copied; this class merely keeps a reference to the original time
    array.  If you change that array *in place* outside of this class, it changes inside of this
    class as well.  You can, of course, change the variable you used to label that array to point to
    some other quantity without affecting the time array stored in this class.

    """

    def __new__(cls, input_array, *args, **kwargs):
        if len(args) > 2:
            raise ValueError("Only one positional argument may be passed")
        if len(args) == 1:
            kwargs["time"] = args[0]
        metadata = copy.copy(getattr(input_array, "_metadata", {}))
        metadata.update(**kwargs)
        input_array = np.asanyarray(input_array).view(complex)
        time = metadata.get("time", None)
        if time is None:
            raise ValueError("Time data must be specified as part of input array or as constructor parameter")
        time = np.asarray(time).view(float)
        if time.ndim != 1:
            raise ValueError(f"Input time array must have exactly 1 dimension; it has {time.ndim}.")
        if input_array.ndim == 0:
            input_array = input_array[np.newaxis, np.newaxis]
        elif input_array.ndim == 1:
            input_array = input_array[np.newaxis, :]
        elif input_array.shape[-2] != time.shape[0] and input_array.shape[-2] != 1:
            raise ValueError(
                "Second-to-last axis of input array must have size 1 or same size as time array.\n            "
                f"Their shapes are {input_array.shape} and {time.shape}, respectively."
            )
        obj = spherical_functions.Modes(input_array, **kwargs).view(cls)
        obj._metadata["time"] = time
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        super().__array_finalize__(obj)
        if "time" not in self._metadata:
            self._metadata["time"] = None

    @property
    def time(self):
        return self._metadata["time"]

    @time.setter
    def time(self, new_time):
        self._metadata["time"][:] = new_time
        return self.time

    @property
    def n_times(self):
        return self.time.size

    u = time

    t = time

    def interpolate(self, new_time, derivative_order=0, out=None):
        new_time = np.asarray(new_time)
        if new_time.ndim != 1:
            raise ValueError(f"New time array must have exactly 1 dimension; it has {new_time.ndim}.")
        new_shape = self.shape[:-2] + (new_time.size, self.shape[-1])
        if out is not None:
            out = np.asarray(out)
            if out.shape != new_shape:
                raise ValueError(
                    f"Output array should have shape {new_shape} for consistency with new time array and modes array"
                )
            if out.dtype != complex:
                raise ValueError(f"Output array should have dtype `complex`; it has dtype {out.dtype}")
        result = out or np.empty(new_shape, dtype=complex)
        if derivative_order > 3:
            raise ValueError(
                f"{type(self)} interpolation uses CubicSpline, and cannot take a derivative of order {derivative_order}"
            )
        spline = CubicSpline(self.u, self.view(np.ndarray), axis=-2)
        if derivative_order < 0:
            spline = spline.antiderivative(-derivative_order)
        elif 0 < derivative_order <= 3:
            spline = spline.derivative(derivative_order)
        result[:] = spline(new_time)
        metadata = self._metadata.copy()
        metadata["time"] = new_time
        return type(self)(result, **metadata)

    def antiderivative(self, antiderivative_order=1):
        """Integrate modes with respect to time"""
        return self.interpolate(self.time, derivative_order=-antiderivative_order)

    def derivative(self, derivative_order=1):
        """Differentiate modes with respect to time"""
        return self.interpolate(self.time, derivative_order=derivative_order)

    @property
    def dot(self):
        """Differentiate modes once with respect to time"""
        return self.derivative()

    @property
    def ddot(self):
        """Differentiate modes twice with respect to time"""
        return self.derivative(2)

    @property
    def int(self):
        """Integrate modes once with respect to time"""
        return self.antiderivative()

    @property
    def iint(self):
        """Integrate modes twice with respect to time"""
        return self.antiderivative(2)

    @property
    def LM(self):
        return spherical_functions.LM_range(self.ell_min, self.ell_max)

    @property
    def eth_GHP(self):
        """Raise spin-weight with GHP convention"""
        return self.eth / np.sqrt(2)

    @property
    def ethbar_GHP(self):
        """Lower spin-weight with GHP convention"""
        return self.ethbar / np.sqrt(2)

    def grid_multiply(self, mts, **kwargs):
        """Compute mode weights of the product of two functions

        This will compute the values of `self` and `mts` on a grid, multiply the grid
        values together, and then return the mode coefficients of the product.  This
        takes less time and memory compared to the `SWSH_modes.Modes.multiply()`
        function, at the risk of introducing aliasing effects if `working_ell_max` is
        too small.

        Parameters
        ----------
        self: ModesTimeSeries
            One of the quantities to multiply.
        mts: ModesTimeSeries
            The quantity to multiply with 'self'.
        working_ell_max: int, optional
            The value of ell_max to be used to define the computation grid. The
            number of theta points and the number of phi points are set to
            2*working_ell_max+1. Defaults to (self.ell_max + mts.ell_max).
        output_ell_max: int, optional
            The value of ell_max in the output mts object. Defaults to self.ell_max.

        """
        import spinsfast
        import spherical_functions as sf
        from spherical_functions import LM_index

        output_ell_max = kwargs.pop("output_ell_max", self.ell_max)
        working_ell_max = kwargs.pop("working_ell_max", self.ell_max + mts.ell_max)
        n_theta = n_phi = 2 * working_ell_max + 1

        if self.n_times != mts.n_times or not np.equal(self.t, mts.t).all():
            raise ValueError("The time series of objects to be multiplied must be the same.")

        # Transform to grid representation
        self_grid = spinsfast.salm2map(
            self.ndarray, self.spin_weight, lmax=self.ell_max, Ntheta=n_theta, Nphi=n_phi
        )
        mts_grid = spinsfast.salm2map(
            mts.ndarray, mts.spin_weight, lmax=mts.ell_max, Ntheta=n_theta, Nphi=n_phi
        )

        product_grid = self_grid * mts_grid
        product_spin_weight = self.spin_weight + mts.spin_weight

        # Transform back to mode representation
        product = spinsfast.map2salm(product_grid, product_spin_weight, lmax=working_ell_max)

        # Convert product ndarray to a ModesTimeSeries object
        product = product[:, : LM_index(output_ell_max, output_ell_max, 0) + 1]
        product = ModesTimeSeries(
            sf.SWSH_modes.Modes(
                product,
                spin_weight=product_spin_weight,
                ell_min=0,
                ell_max=output_ell_max,
                multiplication_truncator=max,
            ),
            time=self.t,
        )
        return product



# Copyright (c) 2018, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/spherical_functions/blob/master/LICENSE>

import warnings
import numpy as np
import spherical_functions as sf
from .. import WaveformModes, Inertial, h


def read_from_h5(file_name, **kwargs):
    """Read data from an H5 file in LVC format"""
    import re
    import h5py
    from scipy.interpolate import InterpolatedUnivariateSpline as Spline

    phase_re = re.compile("phase_l(?P<ell>.*)_m(?P<m>.*)")
    amp_re = re.compile("amp_l(?P<ell>.*)_m(?P<m>.*)")

    with h5py.File(file_name, "r") as f:
        t = f["NRtimes"][:]
        ell_m = np.array(
            [[int(match["ell"]), int(match["m"])] for key in f for match in [phase_re.match(key)] if match]
        )
        ell_min = np.min(ell_m[:, 0])
        ell_max = np.max(ell_m[:, 0])
        data = np.empty((t.size, sf.LM_total_size(ell_min, ell_max)), dtype=complex)
        for ell in range(ell_min, ell_max + 1):
            for m in range(-ell, ell + 1):
                amp = Spline(
                    f[f"amp_l{ell}_m{m}/X"][:], f[f"amp_l{ell}_m{m}/Y"][:], k=int(f[f"amp_l{ell}_m{m}/deg"][()])
                )(t)
                phase = Spline(
                    f[f"phase_l{ell}_m{m}/X"][:], f[f"phase_l{ell}_m{m}/Y"][:], k=int(f[f"phase_l{ell}_m{m}/deg"][()])
                )(t)
                data[:, sf.LM_index(ell, m, ell_min)] = amp * np.exp(1j * phase)
        if "auxiliary-info" in f and "history.txt" in f["auxiliary-info"]:
            history = ("### " + f["auxiliary-info/history.txt"][()].decode().replace("\n", "\n### ")).split("\n")
        else:
            history = [""]
        constructor_statement = f"scri.LVC.read_from_h5('{file_name}')"
        w = WaveformModes(
            t=t,
            data=data,
            ell_min=ell_min,
            ell_max=ell_max,
            frameType=Inertial,
            dataType=h,
            history=history,
            constructor_statement=constructor_statement,
            r_is_scaled_out=True,
            m_is_scaled_out=True,
        )

    return w



"""

Read output with

    import pandas as pd
    pd.read_csv("BMSTransformations.csv")

"""

import sys
from os import walk
from os.path import join, dirname
from fnmatch import filter
import numpy as np
import h5py
from scri.SpEC import estimate_avg_com_motion as eacm


def run_in(filename, i_this, i_tot, f):
    print(f"Working on {filename} -- {i_this} of {i_tot}")
    sys.stdout.flush()
    try:
        with h5py.File(filename, "r") as horizons:
            tf = horizons["AhA.dir/ChristodoulouMass.dat"][-1, 0]
        vals = np.hstack(eacm(filename))
        dir, lev = dirname(filename).split("Lev")
        csv = f"{dir}, {float(lev)}, {', '.join([str(f) for f in np.hstack(vals)])}, {tf}\n"
        f.write(csv)
    except:
        print(f"Failed in {filename} -- {i_this} of {i_tot}")
        sys.stdout.flush()
    print(f"Finished {filename} -- {i_this} of {i_tot}")
    sys.stdout.flush()


if __name__ == "__main__":
    print("Finding files to operate on")
    files = [
        join(root, "Horizons.h5")
        for top_dir in ["Catalog", "Incoming"]
        for root, dirnames, filenames in walk(top_dir)
        for filename in filter(filenames, "Horizons.h5")
    ]
    print(f"Finished finding {len(files)} files to operate on")
    with open("BMSTransformations.csv", "w") as f:
        f.write("dirname,Lev,x0,y0,z0,vx0,vy0,vz0,t0,tf\n")
        for i_this, filename in enumerate(files, 1):
            run_in(filename, i_this, len(files), f)



# Copyright (c) 2021, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

import sxs
from ... import WaveformModes


default_shuffle_widths = sxs.utilities.default_shuffle_widths


def save(w, file_name=None, file_write_mode="w", L2norm_fractional_tolerance=1e-10, log_frame=None, shuffle_widths=default_shuffle_widths):
    """Save a waveform in RPXMB format

    This function is essentially a backwards-compatibility wrapper for the the
    `sxs.rpxmb.save` function; see that function's docstring for details.

    """
    return sxs.rpxmb.save(
        w.to_sxs,
        file_name=file_name,
        file_write_mode=file_write_mode,
        L2norm_fractional_tolerance=L2norm_fractional_tolerance,
        log_frame=log_frame,
        shuffle_widths=shuffle_widths
    )


def load(file_name, ignore_validation=False, check_md5=True, transform_to_inertial=False, **kwargs):
    """Load a waveform in RPXMB format

    This function is essentially a backwards-compatibility wrapper for the the
    `sxs.rpxmb.load` function; see that function's docstring for full details.

    However, note that this function has slightly different behavior (for backwards
    compatibility):

      1. The default value of `ignore_validation` is `False`, which means that a
         `ValueError` is raised whenever validation fails, rather than just a
         warning.

      2. The default value of `transform_to_inertial` is `False`, which means that
         the returned waveform will still be in the corotating frame (specified
         most precisely by the `log_frame` attribute of the returned waveform).

      3. If `transform_to_inertial` is `False`, the return value is a tuple of `(w,
         log_frame)`; otherwise, it is just the waveform `w`.  In both cases,
         `w.log_frame` will contain this data.

    """
    w_sxs = sxs.rpxmb.load(
        file_name,
        ignore_validation=ignore_validation,
        check_md5=check_md5,
        transform_to_inertial=transform_to_inertial,
        **kwargs
    )
    w = WaveformModes.from_sxs(w_sxs)
    w.json_data = w_sxs.json_data
    w.log_frame = w_sxs.log_frame

    if transform_to_inertial:
        return w
    else:
        return w, w.log_frame



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['scri',
 'scri.LVC',
 'scri.SpEC',
 'scri.SpEC.file_io',
 'scri.asymptotic_bondi_data',
 'scri.pn']

package_data = \
{'': ['*'], 'scri.SpEC': ['samples/*']}

install_requires = \
['h5py>=3,<4',
 'numba>=0.55',
 'numpy-quaternion>=2022.4',
 'numpy>=1.20',
 'scipy>=1.0,<2.0',
 'spherical-functions>=2022.4',
 'spinsfast>=2022.4',
 'sxs>=2022.3.3',
 'tqdm>=4.48.2,<4.61.2']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=1.0,<2.0'],
 'mkdocs': ['mkdocs>=1.1.2,<2.0.0'],
 'mktheapidocs': ['mktheapidocs[plugin]>=0.2.0,<0.3.0']}

setup_kwargs = {
    'name': 'scri',
    'version': '2022.3.2',
    'description': 'Time-dependent functions of spin-weighted spherical harmonics',
    'long_description': '[![Test and deploy](https://github.com/moble/scri/actions/workflows/build.yml/badge.svg)](https://github.com/moble/scri/actions/workflows/build.yml)\n[![Documentation Status](https://readthedocs.org/projects/scri/badge/?version=latest)](https://scri.readthedocs.io/en/latest/?badge=latest)\n[![PyPI Version](https://img.shields.io/pypi/v/scri?color=)](https://pypi.org/project/scri/)\n[![Conda Version](https://img.shields.io/conda/vn/conda-forge/scri.svg?color=)](https://anaconda.org/conda-forge/scri)\n[![MIT License](https://img.shields.io/github/license/moble/scri.svg)](https://github.com/moble/scri/blob/main/LICENSE)\n[![DOI](https://zenodo.org/badge/doi/10.5281/zenodo.4041971.svg)](https://doi.org/10.5281/zenodo.4041971)\n\nScri\n====\n\nPython/numba code for manipulating time-dependent functions of spin-weighted\nspherical harmonics\n\n## Citing this code\n\nIf you use this code for academic work (I can\'t actually imagine any other use\nfor it), please cite the latest version that you used in your publication. The DOI is:\n\n* DOI: [10.5281/zenodo.4041972](https://doi.org/10.5281/zenodo.4041972) ([BibTeX entry on Zenodo](https://zenodo.org/record/4041972/export/hx#.YFNpLe1KiV4))\n\nAlso please cite the papers for/by which it was produced:\n\n* "Angular velocity of gravitational radiation from precessing binaries and the\n  corotating frame", Boyle,\n  [Phys. Rev. D, 87, 104006](http://link.aps.org/doi/10.1103/PhysRevD.87.104006)\n  (2013).\n* "Gravitational-wave modes from precessing black-hole binaries", Boyle *et\n  al.*, http://arxiv.org/abs/1409.4431 (2014).\n* "Transformations of asymptotic gravitational-wave data", Boyle,\n  [Phys. Rev. D, 93, 084031](http://link.aps.org/doi/10.1103/PhysRevD.93.084031)\n  (2015).\n\nBibtex entries for these articles can be found\n[here](https://raw.githubusercontent.com/moble/scri/master/scri.bib).  It might\nalso be nice of you to provide a link directly to this source code.\n\n\n## Quick start\n\nAssuming you have the [`anaconda`](http://continuum.io/downloads) distribution\nof python (the preferred distribution for scientific applications),\ninstallation is as simple as\n\n```sh\nconda update -y --all\nconda install -c conda-forge scri\n```\n\nIf you need to install `anaconda` first, it\'s very easy and doesn\'t require root permissions.  Just [download](http://continuum.io/downloads) and follow the instructions — particularly setting your `PATH`.  Also, make sure `PYTHONPATH` and `PYTHONHOME` are *not* set.  Ensure that it worked by running `python --version`.  It should say something about anaconda; if not, you probably forgot to set your `PATH`.  Now just run the installation command above.\n\nThen, in python, you can check to make sure installation worked with\n\n```python\nimport scri\nw = scri.WaveformModes()\n```\n\nNow, `w` is an object to contain time and waveform data, as well as various\nrelated pieces of information -- though it is trivial in this case, because we\nhaven\'t given it any data.  For more information, see the docstrings of `scri`,\n`scri.WaveformModes`, etc.\n\n\n## Dependencies\n\nThe dependencies should be taken care of automatically by the quick\ninstallation instructions above.  However, if you run into problems (or if you\nfoolishly decide not to use anaconda to install things), it may be because you\nare missing some or all of these:\n\n  * Standard packages (come with full anaconda installation)\n    * [`numpy`](http://www.numpy.org/)\n    * [`scipy`](http://scipy.org/)\n    * [`matplotlib`](http://matplotlib.org/)\n    * [`h5py`](http://www.h5py.org/)\n    * [`numba`](http://numba.pydata.org/)\n  * My packages, available from anaconda.org and/or github\n    * [`fftw`](https://github.com/moble/fftw) (not actually mine,\n      but I maintain a copy for easy installation)\n    * [`spinsfast`](https://github.com/moble/spinsfast) (not actually mine,\n      but I maintain a copy with updated python features)\n    * [`quaternion`](https://github.com/moble/quaternion)\n    * [`spherical_functions`](https://github.com/moble/spherical_functions)\n\nAll these dependencies are installed automatically when you use the `conda`\ncommand described above.  The `anaconda` distribution can co-exist with your\nsystem python with no trouble -- you simply add the path to anaconda before\nyour system executables.  In fact, your system python probably needs to stay\ncrusty and old so that your system doesn\'t break, while you want to use a newer\nversion of python to actually run fancy new code like this.  This is what\n`anaconda` does for you.  It installs into your home directory, so it doesn\'t\nrequire root access.  It can be uninstalled easily, since it exists entirely\ninside its own directory.  And updates are trivial.\n\n### "Manual" installation\n\nThe instructions in the "Quick Start" section above should be sufficient, as\nthere really is no good reason not to use `anaconda`.  You will occasionally\nhear people complain about it not working; these people have not installed it\ncorrectly, and have other python-related environment variables that shouldn\'t\nbe there.  You don\'t want to be one of those people.\n\nNonetheless, it is possible to install these packages without anaconda -- in\nprinciple.  The main hurdle to overcome is `numba`.  Maybe there are nice ways\nto install `numba` without `anaconda`.  I don\'t know.  I don\'t care.  But if\nyou\'re awesome enough to do that, you\'re awesome enough to install all the\nother dependencies without advice from me.  But in short, you can either use\nthe `setup.py` files as usual, or just use `pip`:\n\n```sh\npip install git+git://github.com/moble/spinsfast\npip install git+git://github.com/moble/quaternion\npip install git+git://github.com/moble/spherical_functions\npip install git+git://github.com/moble/scri\n```\n\nAnd since you\'re just *soooo* cool, you already know that the `--user` flag is\nmissing from those commands because you\'re presumably using a virtual\nenvironment, hotshot.\n\n(If you\'re really not that cool, and aren\'t using `virtualenv`, you might think\nyou should `sudo` those commands.  But there\'s no need if you just use the\n`--user` flag instead.  That installs packages into your user directory, which\nis usually a better idea.)\n\nNote that `spinsfast` depends (for both building and running) on `fftw`.  If\nyou run into build problems with `spinsfast`, it probably can\'t find the\nheader or library for `fftw`.  See the documentation of my copy of `spinsfast`\n[here](https://github.com/moble/spinsfast#manual-installation) for suggestions\non solving that problem.  Of course, with `conda`, `fftw` is installed in the\nright place from my channel automatically.\n\n\n## Documentation\n\nTutorials and automatically generated API documentation are available on [Read the Docs: scri](https://scri.readthedocs.io/).\n\n## Acknowledgments\n\nThis code is, of course, hosted on github; because it is an open-source\nproject, the hosting is free, and all the wonderful features of github are\navailable, including free wiki space and web page hosting, pull requests, a\nnice interface to the git logs, etc.\n\nEvery change in this code is\n[auomatically tested](https://travis-ci.org/moble/scri) on\n[Travis-CI](https://travis-ci.org/).  This is a free service (for open-source\nprojects like this one), which integrates beautifully with github, detecting\neach commit and automatically re-running the tests.  The code is downloaded and\ninstalled fresh each time, and then tested, on both versions of python (2 and\n3).  This ensures that no change I make to the code breaks either installation\nor any of the features that I have written tests for.\n\nEvery change to this code is also recompiled automatically, bundled into a\n`conda` package, and made available for download from\n[anaconda.org](https://anaconda.org/moble/scri).  Again, because this is an\nopen-source project all those nice features are free.\n\nThe work of creating this code was supported in part by the Sherman Fairchild\nFoundation and by NSF Grants No. PHY-1306125 and AST-1333129.\n',
    'author': 'Michael Boyle',
    'author_email': 'michael.oliver.boyle@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/moble/scri',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
